/*
 * Copyright (c) 2021 CardContact Systems GmbH, Minden, Germany.
 *
 * Redistribution and use in source (source code) and binary (object code)
 * forms, with or without modification, are permitted provided that the
 * following conditions are met:
 * 1. Redistributed source code must retain the above copyright notice, this
 * list of conditions and the disclaimer below.
 * 2. Redistributed object code must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation
 * and/or other materials provided with the distribution.
 * 3. The name of CardContact may not be used to endorse or promote products derived
 * from this software or in any other form without specific prior written
 * permission from CardContact.
 * 4. Redistribution of any modified code must be labeled "Code derived from
 * the original OpenCard Framework".
 *
 * THIS SOFTWARE IS PROVIDED BY CardContact "AS IS" FREE OF CHARGE. CardContact SHALL NOT BE
 * LIABLE FOR INFRINGEMENTS OF THIRD PARTIES RIGHTS BASED ON THIS SOFTWARE.  ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  CardContact DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THIS
 * SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR THAT THE OPERATION OF IT WILL
 * BE UNINTERRUPTED OR ERROR-FREE.  IN NO EVENT, UNLESS REQUIRED BY APPLICABLE
 * LAW, SHALL CardContact BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  ALSO, CardContact IS UNDER NO OBLIGATION
 * TO MAINTAIN, CORRECT, UPDATE, CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS
 * SOFTWARE.
 */

package de.cardcontact.opencard.eac;

import java.security.GeneralSecurityException;
import java.security.InvalidKeyException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PublicKey;
import java.security.Security;
import java.security.Signature;
import java.security.SignatureException;
import java.security.cert.Certificate;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateParsingException;
import java.security.interfaces.ECPublicKey;
import java.security.spec.ECParameterSpec;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.cardcontact.opencard.eac.cvc.AuthenticatedRequest;
import de.cardcontact.opencard.eac.cvc.CVCertificate;
import de.cardcontact.opencard.eac.cvc.CertificateHolderReference;
import de.cardcontact.opencard.eac.cvc.CertificationAuthorityReference;
import de.cardcontact.opencard.eac.cvc.ECPublicKeyTLV;
import de.cardcontact.opencard.eac.cvc.ECSignature;
import de.cardcontact.opencard.eac.cvc.PublicKeyTLV;
import de.cardcontact.tlv.ObjectIdentifier;
import de.cardcontact.tlv.TLV;
import de.cardcontact.tlv.TLVEncodingException;

public class CardVerifiableCertificate extends Certificate {

	final Logger log = LoggerFactory.getLogger(CardVerifiableCertificate.class);

	byte[] encoded;
	AuthenticatedRequest req = null;
	CVCertificate cvc;
	ECParameterSpec domain = null;



	public CardVerifiableCertificate(byte[] certificate) throws CertificateException {
		super("CVC");

		try {
			TLV tlv = TLV.factory(certificate);
			if (tlv.getTag().equals(AuthenticatedRequest.TAG)) {
				this.req = new AuthenticatedRequest(tlv);
				this.cvc = this.req.getCVCertificate();
			} else {
				this.cvc = new CVCertificate(tlv);
			}

			int size = tlv.getSize();
			this.encoded = new byte[size];
			System.arraycopy(certificate, 0, this.encoded, 0, size);
		} catch (TLVEncodingException e) {
			log.error("Decoding CVC failed with ", e);
			throw new CertificateParsingException(e);
		}
	}



	@Override
	public byte[] getEncoded() throws CertificateEncodingException {
		return this.encoded;
	}



	public static String signAlgoForOID(ObjectIdentifier oid) {
		if (ECPublicKeyTLV.id_TA_ECDSA_SHA_256.equals(oid)) {
			return "SHA256withECDSA";
		} else if (ECPublicKeyTLV.id_TA_ECDSA_SHA_384.equals(oid)) {
			return "SHA384withECDSA";
		} else if (ECPublicKeyTLV.id_TA_ECDSA_SHA_512.equals(oid)) {
			return "SHA512withECDSA";
		} else if (ECPublicKeyTLV.id_TA_ECDSA_SHA_224.equals(oid)) {
			return "SHA224withECDSA";
		} else if (ECPublicKeyTLV.id_TA_ECDSA_SHA_1.equals(oid)) {
			return "SHA1withECDSA";
		}
		return null;
	}



	public static String hashAlgoForOID(ObjectIdentifier oid) {
		if (ECPublicKeyTLV.id_TA_ECDSA_SHA_256.equals(oid)) {
			return "SHA-256";
		} else if (ECPublicKeyTLV.id_TA_ECDSA_SHA_384.equals(oid)) {
			return "SHA-384";
		} else if (ECPublicKeyTLV.id_TA_ECDSA_SHA_512.equals(oid)) {
			return "SHA-512";
		} else if (ECPublicKeyTLV.id_TA_ECDSA_SHA_224.equals(oid)) {
			return "SHA-224";
		} else if (ECPublicKeyTLV.id_TA_ECDSA_SHA_1.equals(oid)) {
			return "SHA-1";
		}
		return null;
	}



	@Override
	public void verify(PublicKey key) throws CertificateException, NoSuchAlgorithmException, InvalidKeyException,
			NoSuchProviderException, SignatureException {
		verify(key, (String)null);
	}



	public void verify(PublicKey key, String algo, String sigProvider) throws CertificateException, NoSuchAlgorithmException,
			InvalidKeyException, NoSuchProviderException, SignatureException {
		Signature verifier;

		if ((sigProvider == null) && !"Android Runtime".equals(System.getProperty("java.runtime.name"))) {
			if (Security.getProvider("BC") != null) {
				sigProvider = "BC";
			}
		}

		if (sigProvider == null) {
			verifier = Signature.getInstance(algo);
		} else {
			verifier = Signature.getInstance(algo, sigProvider);
		}

		byte[] wrappedSignature;
		verifier.initVerify(key);
		if (this.req != null) {
			verifier.update(this.cvc.getBytes());
			verifier.update(this.req.getCertificationAuthorityReference().getBytes());
			wrappedSignature = ECSignature.wrapSignature(this.req.getSignatureTLV().getValue());
		} else {
			verifier.update(this.cvc.getCertificateBody().getBytes());
			wrappedSignature = ECSignature.wrapSignature(this.cvc.getSignatureTLV().getValue());
		}

		boolean verified = verifier.verify(wrappedSignature);

		if (!verified) throw new CertificateException("Certificate verification failed.");
		this.domain = ((ECPublicKey)key).getParams();
	}



	@Override
	public void verify(PublicKey key, String sigProvider) throws CertificateException, NoSuchAlgorithmException,
			InvalidKeyException, NoSuchProviderException, SignatureException {
		verify(key, "SHA256withECDSA", sigProvider);
	}



	public String getSignatureAlgorithm() {
		return CardVerifiableCertificate.signAlgoForOID(this.getCVCertificate().getCertificateBody().getPublicKeyTLV().getObjectIdentifier());
	}



	public void verify(CardVerifiableCertificate ca) throws GeneralSecurityException {
		verify(ca.getPublicKey(), ca.getSignatureAlgorithm(), null);
	}



	@Override
	public PublicKey getPublicKey() {
		return this.cvc.getCertificateBody().getPublicKeyTLV().getPublicKey(this.domain);
	}



	public byte[] getCurveOID() {
		if (!(this.cvc.getCertificateBody().getPublicKeyTLV() instanceof ECPublicKeyTLV)) {
			throw new IllegalArgumentException("getCurveOID() only supported for ECPublicKey");
		}
		return ((ECPublicKeyTLV)this.cvc.getCertificateBody().getPublicKeyTLV()).getCurveOID();
	}



	public CertificationAuthorityReference getCertificationAuthorityReference() {
		return this.cvc.getCertificateBody().getCertificationAuthorityReference();
	}



	public CertificationAuthorityReference getOuterCertificationAuthorityReference() {
		return this.req.getCertificationAuthorityReference();
	}



	public CertificateHolderReference getCertificateHolderReference() {
		return this.cvc.getCertificateBody().getCertificateHolderReference();
	}



	public CVCertificate getCVCertificate() {
		return this.cvc;
	}



	public AuthenticatedRequest getAuthenticatedRequest() {
		return this.req;
	}



	public TLV getExtension(Class clazz) {
		return this.cvc.getCertificateBody().getExtensions().getExtension(clazz);
	}



	/**
	 * Return the SubjectPublicKeyIdentifier, which is the SHA-1 hash of the encoded public key
	 *
	 * @return the 20 byte SubjectPublicKeyIdentifier
	 */
	public byte[] getSubjectPublicKeyIdentifier() {

		byte[] spki = null;
		try {
			byte[] hashinp = null;

			PublicKeyTLV puk = this.getCVCertificate().getCertificateBody().getPublicKeyTLV();
			hashinp = puk.getEncodedForSPKI();
			MessageDigest md = MessageDigest.getInstance("SHA1");
			md.update(hashinp, 0, hashinp.length);
			spki = md.digest();
		} catch (Exception e) {
			throw new RuntimeException("getSubjectPublicKeyIdentifier() failed", e);
		}
		return spki;
	}



	@Override
	public String toString() {
		return this.cvc.dump();
	}
}
