/*
 * Copyright (c) 2021 CardContact Systems GmbH, Minden, Germany.
 *
 * Redistribution and use in source (source code) and binary (object code)
 * forms, with or without modification, are permitted provided that the
 * following conditions are met:
 * 1. Redistributed source code must retain the above copyright notice, this
 * list of conditions and the disclaimer below.
 * 2. Redistributed object code must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation
 * and/or other materials provided with the distribution.
 * 3. The name of CardContact may not be used to endorse or promote products derived
 * from this software or in any other form without specific prior written
 * permission from CardContact.
 * 4. Redistribution of any modified code must be labeled "Code derived from
 * the original OpenCard Framework".
 *
 * THIS SOFTWARE IS PROVIDED BY CardContact "AS IS" FREE OF CHARGE. CardContact SHALL NOT BE
 * LIABLE FOR INFRINGEMENTS OF THIRD PARTIES RIGHTS BASED ON THIS SOFTWARE.  ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  CardContact DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THIS
 * SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR THAT THE OPERATION OF IT WILL
 * BE UNINTERRUPTED OR ERROR-FREE.  IN NO EVENT, UNLESS REQUIRED BY APPLICABLE
 * LAW, SHALL CardContact BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  ALSO, CardContact IS UNDER NO OBLIGATION
 * TO MAINTAIN, CORRECT, UPDATE, CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS
 * SOFTWARE.
 */

package de.cardcontact.opencard.eac;

import java.nio.ByteBuffer;
import java.security.GeneralSecurityException;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.security.interfaces.ECPublicKey;
import java.security.spec.ECGenParameterSpec;
import java.util.Arrays;

import javax.crypto.KeyAgreement;
import javax.crypto.Mac;
import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.DESedeKeySpec;
import javax.crypto.spec.SecretKeySpec;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.cardcontact.opencard.eac.cvc.ECPublicKeyTLV;
import de.cardcontact.opencard.eac.cvc.PublicKeyTLV;
import de.cardcontact.opencard.security.IsoSecureChannel;
import de.cardcontact.opencard.security.IsoSecureChannel.SSCPolicyEnum;
import de.cardcontact.tlv.ObjectIdentifier;

public class ChipAuthentication {

	public final static ObjectIdentifier id_CA_ECDH_3DES_CBC_CBC     = new ObjectIdentifier(new int[] { 0,4,0,127,0,7,2,2,3,2,1 });
	public final static ObjectIdentifier id_CA_ECDH_AES_CBC_CMAC_128 = new ObjectIdentifier(new int[] { 0,4,0,127,0,7,2,2,3,2,2 });
	public final static ObjectIdentifier id_CA_ECDH_AES_CBC_CMAC_192 = new ObjectIdentifier(new int[] { 0,4,0,127,0,7,2,2,3,2,3 });
	public final static ObjectIdentifier id_CA_ECDH_AES_CBC_CMAC_256 = new ObjectIdentifier(new int[] { 0,4,0,127,0,7,2,2,3,2,4 });

	final Logger logger = LoggerFactory.getLogger(ChipAuthentication.class);

	StandardizedDomainParameter domainParameter;
	ObjectIdentifier protocol;
	KeyPair caKeyPair;
	SecretKey kenc;
	SecretKey kmac;



	public ChipAuthentication(ObjectIdentifier protocol, StandardizedDomainParameter domainParameter) {
		this.protocol = protocol;
		this.domainParameter = domainParameter;
	}



	public void generateEphemeralCAKeyPair() {
		KeyPairGenerator keyGen = null;
		try {
			keyGen = KeyPairGenerator.getInstance("EC");
			keyGen.initialize(new ECGenParameterSpec(this.domainParameter.name()));
		} catch (NoSuchAlgorithmException|InvalidAlgorithmParameterException e) {
			logger.error("Can not obtain and initialize EC key generator", e);
			throw new RuntimeException(e);
		}

		this.caKeyPair = keyGen.generateKeyPair();
	}



	public byte[] getCompressedPublicKey() {
		ECPublicKey puk = (ECPublicKey)this.caKeyPair.getPublic();

		int keysize = puk.getParams().getCurve().getField().getFieldSize();
		keysize = (keysize + 7) >> 3;

		byte[] comp = new byte[keysize];
		PublicKeyTLV.toUnsignedByteArray(puk.getW().getAffineX(), comp, 0, keysize);
		return comp;
	}



	public byte[] getEncodedPublicKey() {
		ECPublicKey puk = (ECPublicKey)this.caKeyPair.getPublic();

		int keysize = puk.getParams().getCurve().getField().getFieldSize();
		keysize = (keysize + 7) >> 3;

		return ECPublicKeyTLV.encodePoint(puk.getW(), keysize);
	}



	public byte[] getProtocol() {
		return this.protocol.getValue();
	}



	private SecretKey deriveKey(int alg, byte[] sharedSecret, int counter, byte[] nonce) {

		ByteBuffer bb = ByteBuffer.allocate(80);
		bb.put(sharedSecret);
		bb.put(nonce);
		bb.put((byte)0);
		bb.put((byte)0);
		bb.put((byte)0);
		bb.put((byte)counter);

		MessageDigest digest = null;
		try {
			digest = MessageDigest.getInstance(alg <= 2 ? "SHA1" : "SHA256");
		} catch (NoSuchAlgorithmException e) {
			logger.error(e.getMessage(), e);
			throw new RuntimeException(e);
		}
		digest.update(bb.array(), 0, bb.position());
		byte[] md = digest.digest();

		SecretKey key = null;
		if (alg == 1) {
			byte[] keyBin = new byte[24];
			System.arraycopy(md, 0, keyBin, 0, 16);
			System.arraycopy(md, 0, keyBin, 16, 8);

			DESedeKeySpec desedeKeySpec = null;
			try {
				desedeKeySpec = new DESedeKeySpec(keyBin);
			} catch (InvalidKeyException e) {
				logger.error(e.getMessage(), e);
				throw new RuntimeException(e);
			}

			SecretKeyFactory skf = null;
			try {
				skf = SecretKeyFactory.getInstance("DESede");
				key = skf.generateSecret(desedeKeySpec);
			} catch (GeneralSecurityException e) {
				logger.error(e.getMessage(), e);
				throw new RuntimeException(e);
			}
		} else {
			int keylen = 8 * alg;
			byte[] keyBin = new byte[keylen];
			System.arraycopy(md, 0, keyBin, 0, keylen);
			key = new SecretKeySpec(keyBin, "AES");
			Arrays.fill(keyBin, (byte)0);
		}
		Arrays.fill(md, (byte)0);

		return key;
	}



	public void performKeyAgreement(PublicKey peerKey, byte[] nonce) {
		byte[] sharedSecret = null;
		try {
			KeyAgreement ka = KeyAgreement.getInstance("ECDH");
			ka.init(this.caKeyPair.getPrivate());
			ka.doPhase(peerKey, true);
			sharedSecret = ka.generateSecret();
		} catch (GeneralSecurityException e) {
			logger.error("Can not agree key", e);
			throw new RuntimeException(e);
		}

		byte[] oidv = this.protocol.getValue();
		int alg = oidv[oidv.length - 1];  // 1 - 3DES, 2 - AES128, 3 - AES192, 4 - AES256
		kenc = deriveKey(alg, sharedSecret, 1, nonce);
		kmac = deriveKey(alg, sharedSecret, 2, nonce);
		Arrays.fill(sharedSecret, (byte)0);
	}



	public boolean verifyAuthenticationToken(byte[] authToken) {
		byte[] at = null;

		ECPublicKeyTLV pub = new ECPublicKeyTLV(this.protocol, (ECPublicKey)this.caKeyPair.getPublic(), false);

		try {
			Mac mac;
			if (((SecretKeySpec)kmac).getAlgorithm().equals("AES")) {
				mac = Mac.getInstance("AESCMAC");
			} else {
				mac = Mac.getInstance("ISO9797ALG3Mac");
			}
			mac.init(kmac);
			mac.update(pub.getBytes());
			at = mac.doFinal();
			if (at.length > 8) {
				byte[] stripped = new byte[8];
				System.arraycopy(at, 0, stripped, 0, 8);
				at = stripped;
			}
		} catch (GeneralSecurityException e) {
			logger.error(e.getMessage(), e);
			throw new RuntimeException(e);
		}

		return Arrays.equals(at, authToken);
	}



	public IsoSecureChannel getIsoSecureChannel() {
		IsoSecureChannel sc = new IsoSecureChannel();
		sc.setEncKey(kenc);
		sc.setMacKey(kmac);
		if (this.protocol.equals(id_CA_ECDH_3DES_CBC_CBC)) {
			sc.setMACSendSequenceCounter(new byte[8]);
		} else {
			sc.setMACSendSequenceCounter(new byte[16]);
			sc.setSendSequenceCounterPolicy(SSCPolicyEnum.SYNC_AND_ENCRYPT);
		}
		return sc;
	}
}
