/*
 * Copyright (c) 2021 CardContact Systems GmbH, Minden, Germany.
 *
 * Redistribution and use in source (source code) and binary (object code)
 * forms, with or without modification, are permitted provided that the
 * following conditions are met:
 * 1. Redistributed source code must retain the above copyright notice, this
 * list of conditions and the disclaimer below.
 * 2. Redistributed object code must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation
 * and/or other materials provided with the distribution.
 * 3. The name of CardContact may not be used to endorse or promote products derived
 * from this software or in any other form without specific prior written
 * permission from CardContact.
 * 4. Redistribution of any modified code must be labeled "Code derived from
 * the original OpenCard Framework".
 *
 * THIS SOFTWARE IS PROVIDED BY CardContact "AS IS" FREE OF CHARGE. CardContact SHALL NOT BE
 * LIABLE FOR INFRINGEMENTS OF THIRD PARTIES RIGHTS BASED ON THIS SOFTWARE.  ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  CardContact DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THIS
 * SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR THAT THE OPERATION OF IT WILL
 * BE UNINTERRUPTED OR ERROR-FREE.  IN NO EVENT, UNLESS REQUIRED BY APPLICABLE
 * LAW, SHALL CardContact BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  ALSO, CardContact IS UNDER NO OBLIGATION
 * TO MAINTAIN, CORRECT, UPDATE, CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS
 * SOFTWARE.
 */

package de.cardcontact.opencard.eac;

import java.nio.ByteBuffer;
import java.security.PublicKey;
import java.security.cert.CertificateException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.cardcontact.opencard.security.IsoSecureChannelCredential;
import de.cardcontact.opencard.security.SecureChannel;
import de.cardcontact.opencard.service.isocard.IsoCommandAPDU;
import de.cardcontact.opencard.service.isocard.apdu.ChipAuthenticationCommandData;
import de.cardcontact.opencard.service.isocard.apdu.ChipAuthenticationResponseData;
import de.cardcontact.opencard.service.isocard.apdu.GeneralAuthenticateCommandAPDU;
import de.cardcontact.opencard.service.isocard.apdu.GetChallengeCommandAPDU;
import de.cardcontact.opencard.service.isocard.apdu.ManageSECommandAPDU;
import de.cardcontact.opencard.service.isocard.apdu.VerifyCertificateCommandAPDU;
import de.cardcontact.tlv.HexString;
import de.cardcontact.tlv.ObjectIdentifier;
import de.cardcontact.tlv.PrimitiveTLV;
import de.cardcontact.tlv.Sequence;
import de.cardcontact.tlv.TLV;
import de.cardcontact.tlv.TLVEncodingException;
import de.cardcontact.tlv.Tag;
import opencard.core.OpenCardException;
import opencard.core.service.CardChannel;
import opencard.core.service.CardService;
import opencard.core.service.CardServiceException;
import opencard.core.service.CardServiceScheduler;
import opencard.core.service.SmartCard;
import opencard.core.terminal.CommandAPDU;
import opencard.core.terminal.ResponseAPDU;

public class EACCardService extends CardService {

	final Logger logger = LoggerFactory.getLogger(EACCardService.class);

	String rootCHR = null;
	ChipAuthentication ca;
	byte[] idPICC = new byte[0];
	PublicKey caPublicKey;



	@Override
	protected void initialize(CardServiceScheduler scheduler, SmartCard card, boolean blocking) throws CardServiceException {
		super.initialize(scheduler, card, blocking);
	}



	public void setRootCHR(String rootCHR) {
		this.rootCHR = rootCHR;
	}



	public void setChipAuthenticationParameter(ObjectIdentifier protocol, StandardizedDomainParameter domainParameter) {
		this.ca = new ChipAuthentication(protocol, domainParameter);
	}



	public void setChipAuthenticationPublicKey(PublicKey caPublicKey) {
		this.caPublicKey = caPublicKey;
	}



	private CardVerifiableCertificate[] getCVCChain(TerminalAuthenticationSigner tasigner) throws CertificateException  {
		byte chainbin[][] = tasigner.getCertificateChain(this.rootCHR);
		CardVerifiableCertificate[] chain = new CardVerifiableCertificate[chainbin.length];
		for (int i = 0; i < chainbin.length; i++) {
			chain[i] = new CardVerifiableCertificate(chainbin[i]);
		}
		return chain;
	}



	private boolean selectPublicKey(byte[] chr, boolean queueable) throws OpenCardException {
		Sequence tlv = new Sequence();
		tlv.add(new PrimitiveTLV(new Tag(3, Tag.CONTEXT, false), chr));  // Tag 83

		ManageSECommandAPDU cmd = new ManageSECommandAPDU(0x81, 0xB6, tlv);
		cmd.setQueueable(queueable);
		CardChannel channel = getCardChannel();
		ResponseAPDU rsp = channel.sendCommandAPDU(cmd);

		return rsp.sw() == 0x9000;
	}



	private void verifyCertificate(CardVerifiableCertificate cvc) throws OpenCardException {
		VerifyCertificateCommandAPDU cmd = new VerifyCertificateCommandAPDU(cvc);
		cmd.setQueueable(true);
		CardChannel channel = getCardChannel();
		ResponseAPDU rsp = channel.sendCommandAPDU(cmd);
		if (rsp.sw() != 0x9000) {
			throw new CardServiceException("Could not verify certificate (SW=" + Integer.toHexString(rsp.sw()) + ")");
		}
	}



	private void verifyCertificateChain(CardVerifiableCertificate[] chain) throws OpenCardException {
		int i = chain.length - 1;

		while((i >= 0) && !selectPublicKey(chain[i].getCertificationAuthorityReference().getValue(), false)) {
			i--;
		}

		if (i < 0) {
			throw new CardServiceException("Could not select a public key for verification.");
		}

		while(i < chain.length) {
			verifyCertificate(chain[i]);
			i++;
			if (i < chain.length) {
				if (!selectPublicKey(chain[i].getCertificationAuthorityReference().getValue(), true)) {
					throw new CardServiceException("Could not select a public key " + new String(chain[i].getCertificationAuthorityReference().getValue()) + " for verification.");
				}
			}
		}
	}



	public byte[] prepareTerminalAuthentication(ChipAuthentication ca, CardVerifiableCertificate terminalCert, TLV auxdata) throws OpenCardException {
		Sequence tlv = new Sequence();

		ObjectIdentifier taOID = terminalCert.getCVCertificate().getCertificateBody().getPublicKeyTLV().getObjectIdentifier();
		tlv.add(new PrimitiveTLV(new Tag(0, Tag.CONTEXT, false), taOID.getValue()));  // Tag 80 / Algorithm
		byte chr[] = terminalCert.getCertificateHolderReference().getValue();
		tlv.add(new PrimitiveTLV(new Tag(3, Tag.CONTEXT, false), chr));  // Tag 83 / Public Key Reference

		if (auxdata != null) {
			tlv.add(tlv);
		}

		byte idIFD[] = ca.getCompressedPublicKey();
		tlv.add(new PrimitiveTLV(new Tag(17, Tag.CONTEXT, false), idIFD));  // Tag 91 / Compressed CA Public Key

		CommandAPDU cmd = new ManageSECommandAPDU((byte)0x81, (byte)0xA4, tlv);
		CardChannel channel = getCardChannel();
		ResponseAPDU rsp = channel.sendCommandAPDU(cmd);
		if (rsp.sw() != 0x9000) {
			throw new CardServiceException("Could not verify certificate (SW=" + Integer.toHexString(rsp.sw()) + ")");
		}

		cmd = new GetChallengeCommandAPDU(8);
		channel = getCardChannel();
		rsp = channel.sendCommandAPDU(cmd);
		if (rsp.sw() != 0x9000) {
			throw new CardServiceException("Could not obtain challenge");
		}

		ByteBuffer buf = ByteBuffer.allocate(256);
		buf.put(this.idPICC);
		buf.put(rsp.data());
		buf.put(idIFD);
		if (auxdata != null) {
			buf.put(auxdata.getBytes());
		}

		logger.debug("Hash input" + HexString.dump(buf.array(), 0, buf.position()));
		buf.flip();
		byte[] dataTBS = new byte[buf.remaining()];
		buf.get(dataTBS);

		return dataTBS;
	}



	public void performTerminalAuthentication(TerminalAuthenticationSigner tasigner) throws OpenCardException {
		CardVerifiableCertificate[] chain = null;

		try	{
			chain = getCVCChain(tasigner);
		}
		catch(CertificateException ce) {
			logger.error("Certificates returned from TA-Signer invalid", ce);
			throw new CardServiceException("Certificates returned from TA-Signer invalid: " + ce.getMessage());
		}

		this.ca.generateEphemeralCAKeyPair();

		try	{
			allocateCardChannel();
			verifyCertificateChain(chain);

			CardVerifiableCertificate terminalCert = chain[chain.length - 1];
			byte[] hash = prepareTerminalAuthentication(ca, terminalCert, null);
			String chr = new String(terminalCert.getCertificateHolderReference().getValue());
			byte[] signature = tasigner.getTASignature(hash, chr);

			CommandAPDU cmd = new IsoCommandAPDU((byte)0x00, (byte)0x82, (byte)0x00, (byte)0x00, signature);
			CardChannel channel = getCardChannel();
			ResponseAPDU rsp = channel.sendCommandAPDU(cmd);

			if (rsp.sw() != 0x9000) {
				throw new CardServiceException("Terminal authentication failed");
			}
		}
		finally {
			releaseCardChannel();
		}
	}



	private void selectChipAuthenticationParameter() throws OpenCardException {
		Sequence tlv = new Sequence();
		tlv.add(new PrimitiveTLV(new Tag(0, Tag.CONTEXT, false), ca.getProtocol()));  // Tag 80 - Protocol

		ManageSECommandAPDU cmd = new ManageSECommandAPDU(0x41, 0xA4, tlv);
		cmd.setQueueable(true);
		CardChannel channel = getCardChannel();
		ResponseAPDU rsp = channel.sendCommandAPDU(cmd);

		if (rsp.sw() != 0x9000) {
			throw new CardServiceException("Terminal authentication failed");
		}
	}



	public void performChipAuthentication() throws OpenCardException {
		try	{
			allocateCardChannel();
			selectChipAuthenticationParameter();
			ChipAuthenticationCommandData cdata = new ChipAuthenticationCommandData(this.ca.getEncodedPublicKey());

			GeneralAuthenticateCommandAPDU cmd = new GeneralAuthenticateCommandAPDU(cdata);
			CardChannel channel = getCardChannel();
			ResponseAPDU rsp = channel.sendCommandAPDU(cmd);

			if (rsp.sw() != 0x9000) {
				throw new CardServiceException("Chip authentication failed");
			}

			ChipAuthenticationResponseData dad = ChipAuthenticationResponseData.parse(rsp.data());
			this.ca.performKeyAgreement(caPublicKey, dad.getNonce());
			if (!this.ca.verifyAuthenticationToken(dad.getAuthenticationToken()) ) {
				throw new CardServiceException("Authentication token verification failed");
			}
		}
		catch(TLVEncodingException te) {
			logger.error("Failed to parse GENERAL AUTHENTICATE R-Data", te);
			throw new CardServiceException("Failed to parse GENERAL AUTHENTICATE R-Data: " + te.getMessage());
		}
		finally {
			releaseCardChannel();
		}
	}



	public IsoSecureChannelCredential getSecureMessagingCredential() {
		return new IsoSecureChannelCredential(SecureChannel.ALL, this.ca.getIsoSecureChannel());
	}
}
