/*
 * Copyright (c) 2016 CardContact Systems GmbH, Minden, Germany.
 *
 * Redistribution and use in source (source code) and binary (object code)
 * forms, with or without modification, are permitted provided that the
 * following conditions are met:
 * 1. Redistributed source code must retain the above copyright notice, this
 * list of conditions and the disclaimer below.
 * 2. Redistributed object code must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation
 * and/or other materials provided with the distribution.
 * 3. The name of CardContact may not be used to endorse or promote products derived
 * from this software or in any other form without specific prior written
 * permission from CardContact.
 * 4. Redistribution of any modified code must be labeled "Code derived from
 * the original OpenCard Framework".
 *
 * THIS SOFTWARE IS PROVIDED BY CardContact "AS IS" FREE OF CHARGE. CardContact SHALL NOT BE
 * LIABLE FOR INFRINGEMENTS OF THIRD PARTIES RIGHTS BASED ON THIS SOFTWARE.  ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  CardContact DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THIS
 * SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR THAT THE OPERATION OF IT WILL
 * BE UNINTERRUPTED OR ERROR-FREE.  IN NO EVENT, UNLESS REQUIRED BY APPLICABLE
 * LAW, SHALL CardContact BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  ALSO, CardContact IS UNDER NO OBLIGATION
 * TO MAINTAIN, CORRECT, UPDATE, CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS
 * SOFTWARE.
 */

package de.cardcontact.opencard.eac;

import java.util.HashMap;
import java.util.Map;

import de.cardcontact.opencard.eac.cvc.PublicKeyReference;


/**
 * Singleton class providing a process wide trust store of Card Verifiable Certificates (CVC).
 *
 * The trust store is backed by a hashmap that maps the Certificate Holder Reference (CHR) to the
 * certificate. The trust store itself does not validate certificates added to the store. This must be
 * handled by the caller of the addTrustedCertificate() method.
 *
 * @author asc
 */
public class TrustStore {

	private static TrustStore instance = null;
	private Map<String, CardVerifiableCertificate> store = new HashMap<String, CardVerifiableCertificate>();



	/**
	 * Return the singleton instance of the trust store.
	 *
	 * @return the trust store instance
	 */
	public static TrustStore getInstance() {
		if (instance == null) {
			instance = new TrustStore();
		}
		return instance;
	}



	/**
	 * Return true if the trust store is empty.
	 *
	 * @return true if the trust store is empty.
	 */
	public static boolean isEmpty() {
		return instance == null || instance.store.isEmpty();
	}



	/**
	 * Return the CV certificate with the Certificate Holder Reference (CHR) specified
	 * by the id parameter
	 *
	 * @param id the Certificate Holder Reference (CHR) of the requested certificate.
	 * @return The CV certificate or null if no certificate with matching CHR found.
	 */
	public CardVerifiableCertificate getTrustedCertificate(PublicKeyReference id) {
		return store.get(new String(id.getValue()));
	}



	/**
	 * Add a trusted certificate to the trust store.
	 *
	 * The trust store relies on the caller to establish trust in the certificate,
	 * i.e. by checking the validity of the certificate against a certificate already in the
	 * trust store.
	 *
	 * @param cert the certificate to add to the trust store.
	 */
	public void addTrustedCertificate(CardVerifiableCertificate cert) {
		store.put(new String(cert.getCertificateHolderReference().getValue()), cert);
	}
}
