/*
 * Copyright (c) 2021 CardContact Systems GmbH, Minden, Germany.
 *
 * Redistribution and use in source (source code) and binary (object code)
 * forms, with or without modification, are permitted provided that the
 * following conditions are met:
 * 1. Redistributed source code must retain the above copyright notice, this
 * list of conditions and the disclaimer below.
 * 2. Redistributed object code must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation
 * and/or other materials provided with the distribution.
 * 3. The name of CardContact may not be used to endorse or promote products derived
 * from this software or in any other form without specific prior written
 * permission from CardContact.
 * 4. Redistribution of any modified code must be labeled "Code derived from
 * the original OpenCard Framework".
 *
 * THIS SOFTWARE IS PROVIDED BY CardContact "AS IS" FREE OF CHARGE. CardContact SHALL NOT BE
 * LIABLE FOR INFRINGEMENTS OF THIRD PARTIES RIGHTS BASED ON THIS SOFTWARE.  ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  CardContact DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THIS
 * SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR THAT THE OPERATION OF IT WILL
 * BE UNINTERRUPTED OR ERROR-FREE.  IN NO EVENT, UNLESS REQUIRED BY APPLICABLE
 * LAW, SHALL CardContact BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  ALSO, CardContact IS UNDER NO OBLIGATION
 * TO MAINTAIN, CORRECT, UPDATE, CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS
 * SOFTWARE.
 */

package de.cardcontact.opencard.eac.cvc;

import java.security.GeneralSecurityException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.Signature;
import java.security.interfaces.ECPrivateKey;
import java.security.interfaces.ECPublicKey;

import de.cardcontact.tlv.ObjectIdentifier;

public class CVCertificateGenerator {

	private CertificationAuthorityReference car;
	private CertificateHolderReference chr;
	private ObjectIdentifier pubalgo = ECPublicKeyTLV.id_TA_ECDSA_SHA_256;
	private PublicKey pub;
	private CertificateHolderAuthorizationTemplate chat;
	private CertificateEffectiveDate notBefore;
	private CertificateExpirationDate notAfter;
	private Extensions extensions = null;
	private boolean withDomainParameter = true;



	public CVCertificateGenerator() {

	}



	public void withDomainParameter(boolean withDomainParameter) {
		this.withDomainParameter = withDomainParameter;
	}



	public void setCertificationAuthorityReference(CertificationAuthorityReference car) {
		this.car = car;
	}



	public void setPublicKey(PublicKey pub) {
		this.pub = pub;
	}



	public void setPublicKey(PublicKey pub, ObjectIdentifier algo) {
		this.pub = pub;
		this.pubalgo = algo;
	}



	public void setCertificateHolderReference(CertificateHolderReference chr) {
		this.chr = chr;
	}



	public void setCertificateHolderAuthorizationTemplate(CertificateHolderAuthorizationTemplate chat) {
		this.chat = chat;
	}



	public void setCertificateEffectiveDate(CertificateEffectiveDate notBefore) {
		this.notBefore = notBefore;
	}



	public void setCertificateExpirationDate(CertificateExpirationDate notAfter) {
		this.notAfter = notAfter;
	}



	public void addExtension(Extension extension) {
		if (this.extensions == null) {
			this.extensions = new Extensions();
		}
		this.extensions.add(extension);
	}



	public CVCertificate generate(PrivateKey signerkey, String algo) throws GeneralSecurityException {
		PublicKeyTLV pt = new ECPublicKeyTLV(this.pubalgo, (ECPublicKey)this.pub, this.withDomainParameter);
		CertificateBody body = new CertificateBody(car, pt, chr, chat, notBefore, notAfter);

		if (this.extensions != null) {
			body.add(this.extensions);
		}

		Signature signer = Signature.getInstance(algo);
		signer.initSign(signerkey);
		signer.update(body.getBytes());
		byte[] signature = signer.sign();

		if (signerkey instanceof ECPrivateKey) {
			int keysize = ((ECPrivateKey)signerkey).getParams().getCurve().getField().getFieldSize();
			keysize = (keysize + 7) >> 3;
			signature = ECSignature.unwrapSignature(signature, keysize);
		}
		return new CVCertificate(body, new SignatureTLV(signature));
	}



	public CVCertificate generate(PrivateKey signerkey) throws GeneralSecurityException {
		return generate(signerkey, "SHA256withECDSA");
	}
}
