/*
 * Copyright (c) 2021 CardContact Systems GmbH, Minden, Germany.
 *
 * Redistribution and use in source (source code) and binary (object code)
 * forms, with or without modification, are permitted provided that the
 * following conditions are met:
 * 1. Redistributed source code must retain the above copyright notice, this
 * list of conditions and the disclaimer below.
 * 2. Redistributed object code must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation
 * and/or other materials provided with the distribution.
 * 3. The name of CardContact may not be used to endorse or promote products derived
 * from this software or in any other form without specific prior written
 * permission from CardContact.
 * 4. Redistribution of any modified code must be labeled "Code derived from
 * the original OpenCard Framework".
 *
 * THIS SOFTWARE IS PROVIDED BY CardContact "AS IS" FREE OF CHARGE. CardContact SHALL NOT BE
 * LIABLE FOR INFRINGEMENTS OF THIRD PARTIES RIGHTS BASED ON THIS SOFTWARE.  ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  CardContact DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THIS
 * SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR THAT THE OPERATION OF IT WILL
 * BE UNINTERRUPTED OR ERROR-FREE.  IN NO EVENT, UNLESS REQUIRED BY APPLICABLE
 * LAW, SHALL CardContact BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  ALSO, CardContact IS UNDER NO OBLIGATION
 * TO MAINTAIN, CORRECT, UPDATE, CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS
 * SOFTWARE.
 */

package de.cardcontact.opencard.eac.cvc;

import java.util.Arrays;

import de.cardcontact.tlv.ConstructedTLV;
import de.cardcontact.tlv.PrimitiveTLV;
import de.cardcontact.tlv.TLV;
import de.cardcontact.tlv.TLVEncodingException;
import de.cardcontact.tlv.Tag;

public class CertificateBody extends ConstructedTLV {

	final static Tag TAG = new Tag(78, Tag.APPLICATION, true);
	final static String NAME = "Certificate Body";

	final static Tag CPI_TAG = new Tag(41, Tag.APPLICATION, false);
	final static String CPI_NAME = "Certificate Profile Indicator";
	final static byte[] CPI_VALUE = new byte[] { 0x00 };
	boolean isRequest;
	boolean hasCAR;



	public CertificateBody(CertificationAuthorityReference car, PublicKeyTLV pub, CertificateHolderReference chr, CertificateHolderAuthorizationTemplate chat, CertificateEffectiveDate notBefore, CertificateExpirationDate notAfter) {
		super(TAG);
		setName(NAME);
		PrimitiveTLV t = new PrimitiveTLV(CPI_TAG, CPI_VALUE );
		t.setName(CPI_NAME);
		isRequest = false;
		add(t);
		add(car);
		add(pub);
		add(chr);
		add(chat);
		add(notBefore);
		add(notAfter);
	}



	public CertificateBody(CertificationAuthorityReference car, PublicKeyTLV pub, CertificateHolderReference chr) {
		super(TAG);
		setName(NAME);
		PrimitiveTLV t = new PrimitiveTLV(CPI_TAG, CPI_VALUE );
		t.setName(CPI_NAME);
		isRequest = true;
		add(t);
		if (car != null) {
			add(car);
		}
		add(pub);
		add(chr);
	}



	public CertificateBody(TLV tlv) throws TLVEncodingException {
		super(tlv);
		if (!tlv.getTag().equals(TAG)) {
			throw new TLVEncodingException(NAME + " with wrong tag " + tlv.getTag());
		}

		setName(NAME);
		isRequest = childs.size() < 7;

		// Check child range

		int i = 0;
		TLV t = childs.get(i);
		if (!t.getTag().equals(CPI_TAG)) {
			throw new TLVEncodingException(CPI_NAME + " with wrong tag " + t.getTag());
		}
		if (!Arrays.equals(CPI_VALUE, t.getValue())) {
			throw new TLVEncodingException(CPI_NAME + " unsupported version");
		}
		t.setName(CPI_NAME);

		i++;
		if (childs.get(i).getTag().equals(CertificationAuthorityReference.TAG)) {
			childs.set(i, new CertificationAuthorityReference(childs.get(i)));
			hasCAR = true;
			i++;
		}

		childs.set(i, PublicKeyTLV.newPublicKey(childs.get(i)));

		i++;
		childs.set(i, new CertificateHolderReference(childs.get(i)));

		if (!isRequest) {
			i++;
			childs.set(i, new CertificateHolderAuthorizationTemplate(childs.get(i)));

			i++;
			childs.set(i, new CertificateEffectiveDate(childs.get(i)));

			i++;
			childs.set(i, new CertificateExpirationDate(childs.get(i)));
		}

		i++;
		if (i < childs.size()) {
			childs.set(i, new Extensions(childs.get(i)));
		}
	}



	public CertificationAuthorityReference getCertificationAuthorityReference() {
		if (!hasCAR) {
			return null;
		}
		return (CertificationAuthorityReference)childs.get(1);
	}



	public PublicKeyTLV getPublicKeyTLV() {
		return (PublicKeyTLV)childs.get(hasCAR ? 2 : 1);
	}



	public CertificateHolderReference getCertificateHolderReference() {
		return (CertificateHolderReference)childs.get(hasCAR ? 3 : 2);
	}



	public CertificateHolderAuthorizationTemplate getCertificateHolderAuthorizationTemplate() {
		if (isRequest) {
			throw new RuntimeException("Certificate body in request does not have an holder authorization template");
		}
		return (CertificateHolderAuthorizationTemplate)childs.get(hasCAR ? 4 : 3);
	}



	public CertificateEffectiveDate getCertificateEffectiveDate() {
		if (isRequest) {
			throw new RuntimeException("Certificate body in request does not have an effective date");
		}
		return (CertificateEffectiveDate)childs.get(hasCAR ? 5 : 4);
	}



	public CertificateExpirationDate getCertificateExpirationDate() {
		if (isRequest) {
			throw new RuntimeException("Certificate body in request does not have an expiration date");
		}
		return (CertificateExpirationDate)childs.get(hasCAR ? 6 : 5);
	}



	public Extensions getExtensions() {
		int i = (isRequest ? 3 : 6) + (hasCAR ? 1 : 0);
		if (i >= childs.size()) {
			return null;
		}
		return (Extensions)childs.get(i);
	}
}
