/*
 * Copyright (c) 2021 CardContact Systems GmbH, Minden, Germany.
 *
 * Redistribution and use in source (source code) and binary (object code)
 * forms, with or without modification, are permitted provided that the
 * following conditions are met:
 * 1. Redistributed source code must retain the above copyright notice, this
 * list of conditions and the disclaimer below.
 * 2. Redistributed object code must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation
 * and/or other materials provided with the distribution.
 * 3. The name of CardContact may not be used to endorse or promote products derived
 * from this software or in any other form without specific prior written
 * permission from CardContact.
 * 4. Redistribution of any modified code must be labeled "Code derived from
 * the original OpenCard Framework".
 *
 * THIS SOFTWARE IS PROVIDED BY CardContact "AS IS" FREE OF CHARGE. CardContact SHALL NOT BE
 * LIABLE FOR INFRINGEMENTS OF THIRD PARTIES RIGHTS BASED ON THIS SOFTWARE.  ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  CardContact DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THIS
 * SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR THAT THE OPERATION OF IT WILL
 * BE UNINTERRUPTED OR ERROR-FREE.  IN NO EVENT, UNLESS REQUIRED BY APPLICABLE
 * LAW, SHALL CardContact BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  ALSO, CardContact IS UNDER NO OBLIGATION
 * TO MAINTAIN, CORRECT, UPDATE, CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS
 * SOFTWARE.
 */

package de.cardcontact.opencard.eac.cvc;

import java.math.BigInteger;
import java.security.KeyFactory;
import java.security.PublicKey;
import java.security.Security;
import java.security.interfaces.ECPublicKey;
import java.security.spec.AlgorithmParameterSpec;
import java.security.spec.ECField;
import java.security.spec.ECFieldFp;
import java.security.spec.ECParameterSpec;
import java.security.spec.ECPoint;
import java.security.spec.ECPublicKeySpec;
import java.security.spec.EllipticCurve;

import de.cardcontact.tlv.ObjectIdentifier;
import de.cardcontact.tlv.PrimitiveTLV;
import de.cardcontact.tlv.TLV;
import de.cardcontact.tlv.TLVEncodingException;
import de.cardcontact.tlv.Tag;

public class ECPublicKeyTLV extends PublicKeyTLV {

	public final static ObjectIdentifier id_TA_ECDSA_SHA_1 = new ObjectIdentifier(new int[] { 0,4,0,127,0,7,2,2,2,2,1 });
	public final static ObjectIdentifier id_TA_ECDSA_SHA_224 = new ObjectIdentifier(new int[] { 0,4,0,127,0,7,2,2,2,2,2 });
	public final static ObjectIdentifier id_TA_ECDSA_SHA_256 = new ObjectIdentifier(new int[] { 0,4,0,127,0,7,2,2,2,2,3 });
	public final static ObjectIdentifier id_TA_ECDSA_SHA_384 = new ObjectIdentifier(new int[] { 0,4,0,127,0,7,2,2,2,2,4 });
	public final static ObjectIdentifier id_TA_ECDSA_SHA_512 = new ObjectIdentifier(new int[] { 0,4,0,127,0,7,2,2,2,2,5 });

	final static Tag TAG_P = new Tag(1, Tag.CONTEXT, false);
	final static Tag TAG_A = new Tag(2, Tag.CONTEXT, false);
	final static Tag TAG_B = new Tag(3, Tag.CONTEXT, false);
	final static Tag TAG_G = new Tag(4, Tag.CONTEXT, false);
	final static Tag TAG_R = new Tag(5, Tag.CONTEXT, false);
	final static Tag TAG_Y = new Tag(6, Tag.CONTEXT, false);
	final static Tag TAG_F = new Tag(7, Tag.CONTEXT, false);

	final static String NAME_P = "Prime";
	final static String NAME_A = "First Coefficient";
	final static String NAME_B = "Second Coefficient";
	final static String NAME_G = "Base Point";
	final static String NAME_R = "Order of the Base Point";
	final static String NAME_Y = "Public Point";
	final static String NAME_F = "Cofactor";

	public final static byte[] OID_prime192v3 = { (byte)0x2A, (byte)0x86, (byte)0x48, (byte)0xCE, (byte)0x3D, (byte)0x03, (byte)0x01, (byte)0x03 };
	public final static byte[] OID_prime192v2 = { (byte)0x2A, (byte)0x86, (byte)0x48, (byte)0xCE, (byte)0x3D, (byte)0x03, (byte)0x01, (byte)0x02 };
	public final static byte[] OID_prime192v1 = { (byte)0x2A, (byte)0x86, (byte)0x48, (byte)0xCE, (byte)0x3D, (byte)0x03, (byte)0x01, (byte)0x01 };
	public final static byte[] OID_prime256v1 = { (byte)0x2A, (byte)0x86, (byte)0x48, (byte)0xCE, (byte)0x3D, (byte)0x03, (byte)0x01, (byte)0x07 };
	public final static byte[] OID_prime239v3 = { (byte)0x2A, (byte)0x86, (byte)0x48, (byte)0xCE, (byte)0x3D, (byte)0x03, (byte)0x01, (byte)0x06 };
	public final static byte[] OID_prime239v2 = { (byte)0x2A, (byte)0x86, (byte)0x48, (byte)0xCE, (byte)0x3D, (byte)0x03, (byte)0x01, (byte)0x05 };
	public final static byte[] OID_prime239v1 = { (byte)0x2A, (byte)0x86, (byte)0x48, (byte)0xCE, (byte)0x3D, (byte)0x03, (byte)0x01, (byte)0x04 };
	public final static byte[] OID_secp256k1 = { (byte)0x2B, (byte)0x81, (byte)0x04, (byte)0x00, (byte)0x0A };
	public final static byte[] OID_secp521r1 = { (byte)0x2B, (byte)0x81, (byte)0x04, (byte)0x00, (byte)0x23 };
	public final static byte[] OID_secp384r1 = { (byte)0x2B, (byte)0x81, (byte)0x04, (byte)0x00, (byte)0x22 };
	public final static byte[] OID_secp224r1 = { (byte)0x2B, (byte)0x81, (byte)0x04, (byte)0x00, (byte)0x21 };
	public final static byte[] OID_secp192k1 = { (byte)0x2B, (byte)0x81, (byte)0x04, (byte)0x00, (byte)0x1F };
	public final static byte[] OID_brainpoolP224t1 = { (byte)0x2B, (byte)0x24, (byte)0x03, (byte)0x03, (byte)0x02, (byte)0x08, (byte)0x01, (byte)0x01, (byte)0x06 };
	public final static byte[] OID_brainpoolP512t1 = { (byte)0x2B, (byte)0x24, (byte)0x03, (byte)0x03, (byte)0x02, (byte)0x08, (byte)0x01, (byte)0x01, (byte)0x0E };
	public final static byte[] OID_brainpoolP224r1 = { (byte)0x2B, (byte)0x24, (byte)0x03, (byte)0x03, (byte)0x02, (byte)0x08, (byte)0x01, (byte)0x01, (byte)0x05 };
	public final static byte[] OID_brainpoolP512r1 = { (byte)0x2B, (byte)0x24, (byte)0x03, (byte)0x03, (byte)0x02, (byte)0x08, (byte)0x01, (byte)0x01, (byte)0x0D };
	public final static byte[] OID_brainpoolP192t1 = { (byte)0x2B, (byte)0x24, (byte)0x03, (byte)0x03, (byte)0x02, (byte)0x08, (byte)0x01, (byte)0x01, (byte)0x04 };
	public final static byte[] OID_brainpoolP384t1 = { (byte)0x2B, (byte)0x24, (byte)0x03, (byte)0x03, (byte)0x02, (byte)0x08, (byte)0x01, (byte)0x01, (byte)0x0C };
	public final static byte[] OID_brainpoolP192r1 = { (byte)0x2B, (byte)0x24, (byte)0x03, (byte)0x03, (byte)0x02, (byte)0x08, (byte)0x01, (byte)0x01, (byte)0x03 };
	public final static byte[] OID_brainpoolP384r1 = { (byte)0x2B, (byte)0x24, (byte)0x03, (byte)0x03, (byte)0x02, (byte)0x08, (byte)0x01, (byte)0x01, (byte)0x0B };
	public final static byte[] OID_brainpoolP160t1 = { (byte)0x2B, (byte)0x24, (byte)0x03, (byte)0x03, (byte)0x02, (byte)0x08, (byte)0x01, (byte)0x01, (byte)0x02 };
	public final static byte[] OID_brainpoolP320t1 = { (byte)0x2B, (byte)0x24, (byte)0x03, (byte)0x03, (byte)0x02, (byte)0x08, (byte)0x01, (byte)0x01, (byte)0x0A };
	public final static byte[] OID_brainpoolP160r1 = { (byte)0x2B, (byte)0x24, (byte)0x03, (byte)0x03, (byte)0x02, (byte)0x08, (byte)0x01, (byte)0x01, (byte)0x01 };
	public final static byte[] OID_brainpoolP320r1 = { (byte)0x2B, (byte)0x24, (byte)0x03, (byte)0x03, (byte)0x02, (byte)0x08, (byte)0x01, (byte)0x01, (byte)0x09 };
	public final static byte[] OID_brainpoolP256t1 = { (byte)0x2B, (byte)0x24, (byte)0x03, (byte)0x03, (byte)0x02, (byte)0x08, (byte)0x01, (byte)0x01, (byte)0x08 };
	public final static byte[] OID_brainpoolP256r1 = { (byte)0x2B, (byte)0x24, (byte)0x03, (byte)0x03, (byte)0x02, (byte)0x08, (byte)0x01, (byte)0x01, (byte)0x07 };



	public static byte[] encodePoint(ECPoint point, int size) {
		byte[] encoded = new byte[(size << 1) + 1];
		encoded[0] = 0x04;
		PublicKeyTLV.toUnsignedByteArray(point.getAffineX(), encoded, 1, size);
		PublicKeyTLV.toUnsignedByteArray(point.getAffineY(), encoded, 1 + size, size);
		return encoded;
	}



	public static ECPoint decodePoint(byte[] point) {
		int size = (point.length - 1) >> 1;
		byte[] xbin = new byte[size];
		byte[] ybin = new byte[size];
		System.arraycopy(point, 1, xbin, 0, size);
		System.arraycopy(point, 1 + size, ybin, 0, size);

		BigInteger x = new BigInteger(1, xbin);
		BigInteger y = new BigInteger(1, ybin);
		return new ECPoint(x,y);
	}



	public ECPublicKeyTLV(ObjectIdentifier oid, ECPublicKey pk, boolean withDomainParameter) {
		super(oid);
		ECParameterSpec param = pk.getParams();
		EllipticCurve curve = param.getCurve();
		int keysize = curve.getField().getFieldSize();
		keysize = (keysize + 7) >> 3;

		ECFieldFp field = (ECFieldFp)curve.getField();
		TLV t;

		if (withDomainParameter) {
			t = new PrimitiveTLV(TAG_P, PublicKeyTLV.toUnsignedByteArray(field.getP(), keysize));
			t.setName(NAME_P);
			add(t);

			t = new PrimitiveTLV(TAG_A, PublicKeyTLV.toUnsignedByteArray(curve.getA(), keysize));
			t.setName(NAME_A);
			add(t);

			t = new PrimitiveTLV(TAG_B, PublicKeyTLV.toUnsignedByteArray(curve.getB(), keysize));
			t.setName(NAME_B);
			add(t);

			t = new PrimitiveTLV(TAG_G, encodePoint(param.getGenerator(), keysize));
			t.setName(NAME_G);
			add(t);

			t = new PrimitiveTLV(TAG_R, PublicKeyTLV.toUnsignedByteArray(param.getOrder(), keysize));
			t.setName(NAME_R);
			add(t);
		}

		t = new PrimitiveTLV(TAG_Y, encodePoint(pk.getW(), keysize));
		t.setName(NAME_Y);
		add(t);

		if (withDomainParameter) {
			t = new PrimitiveTLV(TAG_F, new byte[] { (byte)param.getCofactor() });
			t.setName(NAME_F);
			add(t);
		}
	}



	public ECPublicKeyTLV(TLV tlv) throws TLVEncodingException {
		super(tlv);

		int size = childs.size();
		if ((size != 2) && (size != 8)) {
			throw new TLVEncodingException(NAME + " must contain 2 or 8 elements");
		}

		TLV t;
		int i = 0;
		if (hasDomainParameter()) {
			i++;
			t = childs.get(i);
			if (!t.getTag().equals(TAG_P)) {
				throw new TLVEncodingException("Expected " + NAME_P + " in " + NAME);
			}
			t.setName(NAME_P);

			i++;
			t = childs.get(i);
			if (!t.getTag().equals(TAG_A)) {
				throw new TLVEncodingException("Expected " + NAME_A + " in " + NAME);
			}
			t.setName(NAME_A);

			i++;
			t = childs.get(i);
			if (!t.getTag().equals(TAG_B)) {
				throw new TLVEncodingException("Expected " + NAME_B + " in " + NAME);
			}
			t.setName(NAME_B);

			i++;
			t = childs.get(i);
			if (!t.getTag().equals(TAG_G)) {
				throw new TLVEncodingException("Expected " + NAME_G + " in " + NAME);
			}
			t.setName(NAME_G);

			i++;
			t = childs.get(i);
			if (!t.getTag().equals(TAG_R)) {
				throw new TLVEncodingException("Expected " + NAME_R + " in " + NAME);
			}
			t.setName(NAME_R);
		}

		i++;
		t = childs.get(i);
		if (!t.getTag().equals(TAG_Y)) {
			throw new TLVEncodingException("Expected " + NAME_Y + " in " + NAME);
		}
		t.setName(NAME_Y);

		if (hasDomainParameter()) {
			i++;
			t = childs.get(i);
			if (!t.getTag().equals(TAG_F)) {
				throw new TLVEncodingException("Expected " + NAME_F + " in " + NAME);
			}
			t.setName(NAME_F);
		}
	}



	public boolean hasDomainParameter() {
		return childs.size() == 8;
	}



	public byte[] getCurveOID() {
		if (!hasDomainParameter()) {
			throw new RuntimeException("EC public key does not have domain parameter");
		}

		byte[] p = childs.get(1).getValue();
		byte[] a = childs.get(2).getValue();
		byte[] b = childs.get(3).getValue();
		int fp = ((p[p.length - 1] & 0xFF) << 16) + ((a[a.length - 1] & 0xFF) << 8) + (b[b.length - 1] & 0xFF);

		switch(fp) {
		case 0xfffc16: return OID_prime192v3;
		case 0xfffc53: return OID_prime192v2;
		case 0xfffcb1: return OID_prime192v1;
		case 0xfffc4b: return OID_prime256v1;
		case 0xfffc3e: return OID_prime239v3;
		case 0xfffc2c: return OID_prime239v2;
		case 0xfffc0a: return OID_prime239v1;
		case 0x2f0007: return OID_secp256k1;
		case 0xfffc00: return OID_secp521r1;
		case 0xfffcef: return OID_secp384r1;
		case 0x01feb4: return OID_secp224r1;
		case 0x370003: return OID_secp192k1;
		case 0xfffc8d: return OID_brainpoolP224t1;
		case 0xf3f03e: return OID_brainpoolP512t1;
		case 0xff430b: return OID_brainpoolP224r1;
		case 0xf3ca23: return OID_brainpoolP512r1;
		case 0x979479: return OID_brainpoolP192t1;
		case 0x5350ee: return OID_brainpoolP384t1;
		case 0x97efc9: return OID_brainpoolP192r1;
		case 0x532611: return OID_brainpoolP384r1;
		case 0x0f0c80: return OID_brainpoolP160t1;
		case 0x272453: return OID_brainpoolP320t1;
		case 0x0f0058: return OID_brainpoolP160r1;
		case 0x27b4a6: return OID_brainpoolP320r1;
		case 0x777404: return OID_brainpoolP256t1;
		case 0x77d9b6: return OID_brainpoolP256r1;
		}
		return null;
	}



	public PublicKey getPublicKey(ECParameterSpec params, String providerName) {
		int i = 6;
		if (hasDomainParameter()) {
			ECField field = new ECFieldFp(new BigInteger(1, childs.get(1).getValue()));
			EllipticCurve curve = new EllipticCurve(field,
					new BigInteger(1, childs.get(2).getValue()),
					new BigInteger(1, childs.get(3).getValue()));


			ECPoint g = decodePoint(childs.get(4).getValue());
			BigInteger r = new BigInteger(1, childs.get(5).getValue());
			int f = childs.get(7).getValue()[0];
			params = new ECParameterSpec(curve, g, r, f);
		} else {
			if (params == null) {
				throw new RuntimeException("EC public key does not have domain parameter");
			}
			i = 1;
		}

		ECPoint y = decodePoint(childs.get(i).getValue());
		ECPublicKeySpec spec = new ECPublicKeySpec(y, params);

		KeyFactory fact = null;
		PublicKey key = null;

		try	{
			if ((providerName == null) && !"Android Runtime".equals(System.getProperty("java.runtime.name"))) {
				if (Security.getProvider("BC") != null) {
					providerName = "BC";
				}
			}

			if (providerName != null) {
				fact = KeyFactory.getInstance("EC", providerName);
			} else {
				fact = KeyFactory.getInstance("EC");
			}
			key = fact.generatePublic(spec);
		}
		catch(RuntimeException e) {
			throw e;
		}
		catch(Exception e) {
			throw new RuntimeException("Problem creating public key", e);
		}

		return key;
	}



	@Override
	public PublicKey getPublicKey(AlgorithmParameterSpec params) {
		return getPublicKey((ECParameterSpec)params, null);
	}



	public byte[] getEncodedForSPKI() {
		int i = hasDomainParameter() ? 6 : 1;
		return childs.get(i).getValue();
	}
}
