/*
 * Copyright (c) 2021 CardContact Systems GmbH, Minden, Germany.
 *
 * Redistribution and use in source (source code) and binary (object code)
 * forms, with or without modification, are permitted provided that the
 * following conditions are met:
 * 1. Redistributed source code must retain the above copyright notice, this
 * list of conditions and the disclaimer below.
 * 2. Redistributed object code must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation
 * and/or other materials provided with the distribution.
 * 3. The name of CardContact may not be used to endorse or promote products derived
 * from this software or in any other form without specific prior written
 * permission from CardContact.
 * 4. Redistribution of any modified code must be labeled "Code derived from
 * the original OpenCard Framework".
 *
 * THIS SOFTWARE IS PROVIDED BY CardContact "AS IS" FREE OF CHARGE. CardContact SHALL NOT BE
 * LIABLE FOR INFRINGEMENTS OF THIRD PARTIES RIGHTS BASED ON THIS SOFTWARE.  ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  CardContact DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THIS
 * SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR THAT THE OPERATION OF IT WILL
 * BE UNINTERRUPTED OR ERROR-FREE.  IN NO EVENT, UNLESS REQUIRED BY APPLICABLE
 * LAW, SHALL CardContact BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  ALSO, CardContact IS UNDER NO OBLIGATION
 * TO MAINTAIN, CORRECT, UPDATE, CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS
 * SOFTWARE.
 */

package de.cardcontact.opencard.eac.cvc;

import java.math.BigInteger;

import de.cardcontact.tlv.IntegerTLV;
import de.cardcontact.tlv.Sequence;
import de.cardcontact.tlv.TLV;
import de.cardcontact.tlv.TLVEncodingException;

public class ECSignature {


	/**
	 * Wrap an ECDSA signature in the format r || s into a TLV encoding as defined by RFC 3279
	 *
	 * @param signature containing the concatenation of r and s as unsigned integer values
	 * @return ASN.1 SEQUENCE objects containing two signed integer r and s
	 * @throws TLVEncodingException
	 */
	public static byte[] wrapSignature(byte[] signature) {
		int length = signature.length / 2;

		byte ib[] = new byte[length];
		System.arraycopy(signature, 0, ib, 0, length);
		BigInteger r = new BigInteger(1, ib);

		System.arraycopy(signature, length, ib, 0, length);
		BigInteger s = new BigInteger(1, ib);

		Sequence sequence = new Sequence();
		sequence.add(new IntegerTLV(r));
		sequence.add(new IntegerTLV(s));

		return sequence.getBytes();
	}



	public static byte[] unwrapSignature(byte[] signature, int size) {
		TLV t;

		try {
			t = TLV.factory(signature);
		} catch (TLVEncodingException e) {
			throw new IllegalArgumentException("Signature is not a TLV object");
		}

		if (!(t instanceof Sequence)) {
			throw new IllegalArgumentException("Signature missing outer Sequence");
		}

		Sequence s = (Sequence)t;
		if (s.getChildCount() != 2) {
			throw new IllegalArgumentException("Signature must contain two elements");
		}

		if (!(s.getChildAt(0) instanceof IntegerTLV) || !(s.getChildAt(1) instanceof IntegerTLV)) {
			throw new IllegalArgumentException("Signature must contain two integer");
		}

		byte[] uwsig = new byte[size << 1];
		IntegerTLV i = (IntegerTLV)s.getChildAt(0);
		PublicKeyTLV.toUnsignedByteArray(i.getValue(), uwsig, 0, size);

		i = (IntegerTLV)s.getChildAt(1);
		PublicKeyTLV.toUnsignedByteArray(i.getValue(), uwsig, size, size);

		return uwsig;
	}



	/**
	 * Rewrap a signature that contains redundant 00 bytes in the encoded integer
	 *
	 * @param signature the TLV encoded signature
	 * @return the rewrapped signature
	 */
	public static byte[] rewrapSignature(byte[] signature) {
		TLV t;

		try {
			t = TLV.factory(signature);
		} catch (TLVEncodingException e) {
			throw new IllegalArgumentException("Signature is not a TLV object");
		}

		if (!(t instanceof Sequence)) {
			throw new IllegalArgumentException("Signature missing outer Sequence");
		}

		Sequence seq = (Sequence)t;
		if (seq.getChildCount() != 2) {
			throw new IllegalArgumentException("Signature must contain two elements");
		}

		if (!(seq.getChildAt(0) instanceof IntegerTLV) || !(seq.getChildAt(1) instanceof IntegerTLV)) {
			throw new IllegalArgumentException("Signature must contain two integer");
		}

		IntegerTLV r = (IntegerTLV)seq.getChildAt(0);
		IntegerTLV s = (IntegerTLV)seq.getChildAt(1);

		if (((r.getValue()[0] == 0) && (r.getValue()[1] >= 0)) ||
			((s.getValue()[0] == 0) && (s.getValue()[1] >= 0))) {
			seq = new Sequence();
			seq.add(new IntegerTLV(r.getBigInteger()));
			seq.add(new IntegerTLV(s.getBigInteger()));

			return seq.getBytes();
		}

		return signature;
	}
}
