/*
 * Copyright (c) 2021 CardContact Systems GmbH, Minden, Germany.
 *
 * Redistribution and use in source (source code) and binary (object code)
 * forms, with or without modification, are permitted provided that the
 * following conditions are met:
 * 1. Redistributed source code must retain the above copyright notice, this
 * list of conditions and the disclaimer below.
 * 2. Redistributed object code must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation
 * and/or other materials provided with the distribution.
 * 3. The name of CardContact may not be used to endorse or promote products derived
 * from this software or in any other form without specific prior written
 * permission from CardContact.
 * 4. Redistribution of any modified code must be labeled "Code derived from
 * the original OpenCard Framework".
 *
 * THIS SOFTWARE IS PROVIDED BY CardContact "AS IS" FREE OF CHARGE. CardContact SHALL NOT BE
 * LIABLE FOR INFRINGEMENTS OF THIRD PARTIES RIGHTS BASED ON THIS SOFTWARE.  ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  CardContact DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THIS
 * SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR THAT THE OPERATION OF IT WILL
 * BE UNINTERRUPTED OR ERROR-FREE.  IN NO EVENT, UNLESS REQUIRED BY APPLICABLE
 * LAW, SHALL CardContact BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  ALSO, CardContact IS UNDER NO OBLIGATION
 * TO MAINTAIN, CORRECT, UPDATE, CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS
 * SOFTWARE.
 */

package de.cardcontact.opencard.eac.cvc;

import java.math.BigInteger;
import java.security.KeyFactory;
import java.security.PublicKey;
import java.security.interfaces.RSAPublicKey;
import java.security.spec.AlgorithmParameterSpec;
import java.security.spec.RSAPublicKeySpec;

import de.cardcontact.tlv.IntegerTLV;
import de.cardcontact.tlv.ObjectIdentifier;
import de.cardcontact.tlv.PrimitiveTLV;
import de.cardcontact.tlv.Sequence;
import de.cardcontact.tlv.TLV;
import de.cardcontact.tlv.TLVEncodingException;
import de.cardcontact.tlv.Tag;

public class RSAPublicKeyTLV extends PublicKeyTLV {

	final static Tag TAG_MODULUS = new Tag(1, Tag.CONTEXT, false);
	final static Tag TAG_EXPONENT = new Tag(2, Tag.CONTEXT, false);

	final static String NAME_MODULUS = "Modulus";
	final static String NAME_EXPONENT = "Public Exponent";


	public RSAPublicKeyTLV(ObjectIdentifier oid, RSAPublicKey pk) {
		super(oid);

		TLV t = new PrimitiveTLV(TAG_MODULUS, pk.getModulus().toByteArray());
		t.setName(NAME_MODULUS);
		add(t);

		t = new PrimitiveTLV(TAG_EXPONENT, pk.getPublicExponent().toByteArray());
		t.setName(NAME_MODULUS);
		add(t);
	}



	public RSAPublicKeyTLV(TLV tlv) throws TLVEncodingException {
		super(tlv);

		if (childs.size() != 3) {
			throw new TLVEncodingException(NAME + " must contain 3 elements");
		}

		TLV t;
		int i = 1;
		t = childs.get(i);
		if (!t.getTag().equals(TAG_MODULUS)) {
			throw new TLVEncodingException("Expected " + NAME_MODULUS + " in " + NAME);
		}
		t.setName(NAME_MODULUS);

		i++;
		t = childs.get(i);
		if (!t.getTag().equals(TAG_EXPONENT)) {
			throw new TLVEncodingException("Expected " + NAME_EXPONENT + " in " + NAME);
		}
		t.setName(NAME_EXPONENT);
	}



	public PublicKey getPublicKey(String providerName) {
		BigInteger modulus = new BigInteger(1, childs.get(1).getValue());
		BigInteger exponent = new BigInteger(1, childs.get(2).getValue());
		RSAPublicKeySpec spec = new RSAPublicKeySpec(modulus, exponent);

		KeyFactory fact = null;
		PublicKey key = null;

		try	{
			if (providerName != null) {
				fact = KeyFactory.getInstance("RSA", providerName);
			} else {
				fact = KeyFactory.getInstance("RSA");
			}
			key = fact.generatePublic(spec);
		}
		catch(RuntimeException e) {
			throw e;
		}
		catch(Exception e) {
			throw new RuntimeException("Problem creating public key", e);
		}

		return key;
	}



	@Override
	public PublicKey getPublicKey(AlgorithmParameterSpec params) {
		return getPublicKey((String)null);
	}



	@Override
	public byte[] getEncodedForSPKI() {
		BigInteger modulus = new BigInteger(1, childs.get(1).getValue());
		BigInteger exponent = new BigInteger(1, childs.get(2).getValue());
		Sequence seq = new Sequence();
		seq.add(new IntegerTLV(modulus));
		seq.add(new IntegerTLV(exponent));
		return seq.getBytes();
	}
}
