/*
 * Copyright (c) 2021 CardContact Systems GmbH, Minden, Germany.
 *
 * Redistribution and use in source (source code) and binary (object code)
 * forms, with or without modification, are permitted provided that the
 * following conditions are met:
 * 1. Redistributed source code must retain the above copyright notice, this
 * list of conditions and the disclaimer below.
 * 2. Redistributed object code must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation
 * and/or other materials provided with the distribution.
 * 3. The name of CardContact may not be used to endorse or promote products derived
 * from this software or in any other form without specific prior written
 * permission from CardContact.
 * 4. Redistribution of any modified code must be labeled "Code derived from
 * the original OpenCard Framework".
 *
 * THIS SOFTWARE IS PROVIDED BY CardContact "AS IS" FREE OF CHARGE. CardContact SHALL NOT BE
 * LIABLE FOR INFRINGEMENTS OF THIRD PARTIES RIGHTS BASED ON THIS SOFTWARE.  ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  CardContact DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THIS
 * SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR THAT THE OPERATION OF IT WILL
 * BE UNINTERRUPTED OR ERROR-FREE.  IN NO EVENT, UNLESS REQUIRED BY APPLICABLE
 * LAW, SHALL CardContact BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  ALSO, CardContact IS UNDER NO OBLIGATION
 * TO MAINTAIN, CORRECT, UPDATE, CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS
 * SOFTWARE.
 */

package de.cardcontact.opencard.eac.softTA;

import java.io.File;
import java.io.FileInputStream;
import java.security.GeneralSecurityException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.Signature;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateException;
import java.security.interfaces.ECPrivateKey;
import java.util.ArrayList;

import de.cardcontact.opencard.eac.CardVerifiableCertificate;
import de.cardcontact.opencard.eac.PKCS8;
import de.cardcontact.opencard.eac.TerminalAuthenticationSigner;
import de.cardcontact.opencard.eac.cvc.ECSignature;

public class SoftTASigner implements TerminalAuthenticationSigner {

	String cspath;
	String signerpath;
	String signerCHR;


	ArrayList<CardVerifiableCertificate> chain = new ArrayList<CardVerifiableCertificate>();
	PrivateKey signer;



	public SoftTASigner(String cspath, String signerpath) {
		this.cspath = cspath;
		this.signerpath = signerpath;
		try	{
			setup();
		}
		catch(Exception e) {
			throw new RuntimeException("Problem setting up signer", e);
		}
	}



	private byte[] loadBinary(String filename) {
		File f = new File(filename);
		if (!f.exists()) {
			throw new RuntimeException("File " + filename + " does not exist");
		}
		int len = (int)(f.length());
		byte[] bin = new byte[len];

		try {
			FileInputStream fis = new FileInputStream(f);
			fis.read(bin);
			fis.close();
		} catch (Exception e) {
			throw new RuntimeException("File " + filename + " could not be read", e);
		}
		return bin;
	}



	private void loadCertificate(String path, String chr) throws CertificateException {
		String filename = this.cspath + path + "/" + chr + ".cvcert";
		File f = new File(filename);
		if (!f.exists()) {
			filename = this.cspath + path + "/" + chr + ".selfsigned.cvcert";
		}

		byte[] bin = loadBinary(filename);
		CardVerifiableCertificate cvc = new CardVerifiableCertificate(bin);

		chain.add(0, cvc);
	}



	private void loadPrivateKey(String path, String chr) throws Exception {
		String filename = this.cspath + path + "/" + chr + ".pkcs8";
		byte[] bin = loadBinary(filename);

		/*
		PKCS8EncodedKeySpec spec = new PKCS8EncodedKeySpec(bin);
		KeyFactory factory = KeyFactory.getInstance("EC");
		this.signer = factory.generatePrivate(spec);
		*/

		this.signer = PKCS8.decodePrivateKey(bin);
	}



	private void setup() throws Exception {
		String seq = "00001";
		String[] elem = this.signerpath.substring(1).split("/");

		String path = this.signerpath;
		String chr = elem[elem.length - 1] + seq;
		loadCertificate(path, chr);

		path = path.substring(0, path.lastIndexOf('/'));
		String car = new String(chain.get(0).getCertificationAuthorityReference().getValue());
		loadCertificate(path, car);

		path = path.substring(0, path.lastIndexOf('/'));
		car = new String(chain.get(0).getCertificationAuthorityReference().getValue());
		loadCertificate(path, car);

		car = new String(chain.get(0).getCertificationAuthorityReference().getValue());
		while (!car.equals(new String(chain.get(0).getCertificateHolderReference().getValue()))) {
			loadCertificate(path, car);
			car = new String(chain.get(0).getCertificationAuthorityReference().getValue());
		}

		loadPrivateKey(this.signerpath, chr);

		CardVerifiableCertificate cvc = chain.get(0);
		// String algo = CardVerifiableCertificate.algoForOID(root.getTA);
		String algo = "SHA256withECDSA";
		PublicKey pub = cvc.getPublicKey();
		for (int i = 1; i < chain.size(); i++) {
			cvc = chain.get(i);
			cvc.verify(pub, algo, null);
			pub = cvc.getPublicKey();
		}

		this.signerCHR = new String(cvc.getCertificateHolderReference().getValue());
		byte[] hash = new byte[] { 0,1,2,3,4,5,6,7,8,9,0,1,2,3,4,5,6,7,8,9,0,1,2,3,4,5,6,7,8,9,0,1 };
		algo = "NONEwithECDSA";
		Signature signer = Signature.getInstance(algo);
		signer.initSign(this.signer);
		signer.update(hash);
		byte[] res = signer.sign();

		signer = Signature.getInstance(algo);
		signer.initVerify(pub);
		signer.update(hash);
		boolean ok = signer.verify(res);
		if (!ok) {
			throw new GeneralSecurityException("Signer validation failed. Private key matches certificate ?");
		}
	}



	@Override
	public byte[][] getCertificateChain(String keyCAR) {
		int i = chain.size() - 1;
		for (; i >= 0; i--) {
			String car = new String(chain.get(i).getCertificationAuthorityReference().getValue());
			if (car.equals(keyCAR)) {
				break;
			}
		}

		if (i < 0) {
			throw new RuntimeException("No certificate along the chain found that can be verified with " + keyCAR);
		}

		int s = chain.size() - i;
		byte[][] res = new byte[s][];
		for (int j = 0; j < s; j++) {
			byte[] enc = null;
			try	{
				enc = chain.get(i + j).getEncoded();
			}
			catch(CertificateEncodingException e) {
				// Ignore: Can not happen with CVCs
			}
			res[j] = enc;
		}

		return res;
	}



	@Override
	public byte[] getTASignature(byte[] dataTBS, String keyCHR) {
		if (!keyCHR.equals(this.signerCHR)) {
			throw new RuntimeException("Key " + keyCHR + " not found");
		}

		byte[] signature = null;
		CardVerifiableCertificate cvc = chain.get(0);
		String algo = cvc.getSignatureAlgorithm();

		try	{
			Signature signer = Signature.getInstance(algo);
			signer.initSign(this.signer);
			signer.update(dataTBS);
			signature = signer.sign();

			if (this.signer instanceof ECPrivateKey) {
				int keysize = ((ECPrivateKey)this.signer).getParams().getCurve().getField().getFieldSize();
				keysize = (keysize + 7) >> 3;
				signature = ECSignature.unwrapSignature(signature, keysize);
			}
		}
		catch(GeneralSecurityException e) {
			throw new RuntimeException("Problem signing challenge", e);
		}

		return signature;
	}
}
