/*
 * Copyright (c) 2016 CardContact Systems GmbH, Minden, Germany.
 *
 * Redistribution and use in source (source code) and binary (object code)
 * forms, with or without modification, are permitted provided that the
 * following conditions are met:
 * 1. Redistributed source code must retain the above copyright notice, this
 * list of conditions and the disclaimer below.
 * 2. Redistributed object code must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation
 * and/or other materials provided with the distribution.
 * 3. The name of CardContact may not be used to endorse or promote products derived
 * from this software or in any other form without specific prior written
 * permission from CardContact.
 * 4. Redistribution of any modified code must be labeled "Code derived from
 * the original OpenCard Framework".
 *
 * THIS SOFTWARE IS PROVIDED BY CardContact "AS IS" FREE OF CHARGE. CardContact SHALL NOT BE
 * LIABLE FOR INFRINGEMENTS OF THIRD PARTIES RIGHTS BASED ON THIS SOFTWARE.  ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  CardContact DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THIS
 * SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR THAT THE OPERATION OF IT WILL
 * BE UNINTERRUPTED OR ERROR-FREE.  IN NO EVENT, UNLESS REQUIRED BY APPLICABLE
 * LAW, SHALL CardContact BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  ALSO, CardContact IS UNDER NO OBLIGATION
 * TO MAINTAIN, CORRECT, UPDATE, CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS
 * SOFTWARE.
 */

package de.cardcontact.opencard.security;

import java.security.GeneralSecurityException;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.util.Arrays;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.Mac;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;



/**
 * Message Authentication Code generator supporting a fallback CMAC (e.g. for Android)
 *
 */
public abstract class MessageAuthenticationCode {

	public static MessageAuthenticationCode getInstance(String algo, String provider) throws GeneralSecurityException {
		MessageAuthenticationCode mac;

		try	{
			mac = new MessageAuthenticationCode.NativeMAC(algo, provider);
		}
		catch(Exception e) {
			mac = new MessageAuthenticationCode.CBCBasedCMAC(algo, provider);
		}
		return mac;
	}



	public abstract void init(Key key);



	public abstract byte[] doFinal(byte[] input);



	private static class NativeMAC extends MessageAuthenticationCode {

		Mac mac;

		public NativeMAC(String algo, String provider) throws NoSuchAlgorithmException, NoSuchProviderException {
			if (provider != null) {
				mac = Mac.getInstance(algo, provider);
			} else {
				mac = Mac.getInstance(algo);
			}
		}



		@Override
		public void init(Key key) {
			try {
				mac.init(key);
			} catch (InvalidKeyException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}



		@Override
		public byte[] doFinal(byte[] input) {
			return mac.doFinal(input);
		}
	}



	private static class CBCBasedCMAC extends MessageAuthenticationCode {

		private Cipher aes;
		private final static byte[] zero = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
		private final static byte[] rb = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, (byte)0x87};
		private final static IvParameterSpec iv = new IvParameterSpec(zero);
		private byte[] k1 = new byte[16];
		private byte[] k2 = new byte[16];

		public CBCBasedCMAC(String algo, String provider) {
			try {
				if (provider != null) {
					aes = Cipher.getInstance("AES/CBC/NoPadding", provider);
				} else {
					aes = Cipher.getInstance("AES/CBC/NoPadding");
				}
			} catch (Exception e) {
				throw new RuntimeException("Could not create AES/CBC/NoPadding instance", e);
			}
		}



		@Override
		public void init(Key key) {
			try {
				aes.init(Cipher.ENCRYPT_MODE, key, iv);
				byte[] k0 = aes.doFinal(zero);

				byte s = (byte)(k0[0] & 0x80);
				doLeftShiftOneBit(k0);
				if (s == (byte)0x80) {
					xor128(k0, 0, rb);
				}
				System.arraycopy(k0, 0, k1, 0, 16);

				s = (byte)(k0[0] & 0x80);
				doLeftShiftOneBit(k0);
				if (s == (byte)0x80) {
					xor128(k0, 0, rb);
				}
				System.arraycopy(k0, 0, k2, 0, 16);

			} catch (InvalidKeyException | IllegalBlockSizeException | BadPaddingException | InvalidAlgorithmParameterException e) {
				throw new RuntimeException("Could not set AES key and calculate k0", e);
			}
		}



		@Override
		public byte[] doFinal(byte[] input) {
			int size = ((input.length + 15) / 16) * 16;
			if (size == 0) {
				size = 16;
			}
			byte[] buf = new byte[size];
			System.arraycopy(input, 0, buf, 0, input.length);

			if (size == input.length) {
				xor128(buf, size - 16, k1);
			} else {
				buf[input.length] = (byte)0x80;
				xor128(buf, size - 16, k2);
			}

			byte[] res;
			try {
				res = aes.doFinal(buf);
			} catch (IllegalBlockSizeException | BadPaddingException e) {
				throw new RuntimeException("Could not perform AES/CBC", e);
			}

			byte[] mac = new byte[16];
			System.arraycopy(res, size - 16, mac, 0, 16);
			return mac;
		}



		/*
		 * XOR two blocks of identical length
		 */
		private static void xor128(byte[] buf1, int offsetBuf1, byte[] buf2) {
			for (int i = 0; i < buf2.length; i++) {
				buf1[offsetBuf1 + i] = (byte)(buf1[offsetBuf1 + i] ^ buf2[i]);
			}
		}



		/*
		 * Shift the complete buffer one bit to the left (MSB first)
		 */
		private static void doLeftShiftOneBit(byte[] buf) {
			byte d;
			byte overflow = 0;

			for(int i = buf.length - 1; i >= 0; i--) {
				d = (byte)((buf[i] << 1) | overflow);
				overflow = ((buf[i] & 0x80) == 0x80) ? (byte)1 : (byte)0; // MSB was 1
				buf[i] = d;
			}
		}



		public static void test() {
			byte[] k = new byte[] { 0x2b,(byte)0x7e,(byte)0x15,(byte)0x16,(byte)0x28,(byte)0xae,(byte)0xd2,(byte)0xa6,(byte)0xab,(byte)0xf7,(byte)0x15,(byte)0x88,(byte)0x09,(byte)0xcf,(byte)0x4f,(byte)0x3c };
			Key key = new SecretKeySpec(k, "AES");
			CBCBasedCMAC m = new CBCBasedCMAC("", "");
			m.init(key);
			byte[] msg = new byte[0];
			byte[] ref = new byte[] { (byte)0xbb,(byte)0x1d,(byte)0x69,(byte)0x29,(byte)0xe9,(byte)0x59,(byte)0x37,(byte)0x28,(byte)0x7f,(byte)0xa3,(byte)0x7d,(byte)0x12,(byte)0x9b,(byte)0x75,(byte)0x67,(byte)0x46 };
			byte[] mac = m.doFinal(msg);
			if (!Arrays.equals(ref, mac)) {
				throw new RuntimeException("AES CMAC Selftest failed.");
			}

			msg = new byte[] { (byte)0x6b,(byte)0xc1,(byte)0xbe,(byte)0xe2,(byte)0x2e,(byte)0x40,(byte)0x9f,(byte)0x96,(byte)0xe9,(byte)0x3d,(byte)0x7e,(byte)0x11,(byte)0x73,(byte)0x93,(byte)0x17,(byte)0x2a };
			ref = new byte[] { (byte)0x07,(byte)0x0a,(byte)0x16,(byte)0xb4,(byte)0x6b,(byte)0x4d,(byte)0x41,(byte)0x44,(byte)0xf7,(byte)0x9b,(byte)0xdd,(byte)0x9d,(byte)0xd0,(byte)0x4a,(byte)0x28,(byte)0x7c };
			mac = m.doFinal(msg);
			if (!Arrays.equals(ref, mac)) {
				throw new RuntimeException("AES CMAC Selftest failed.");
			}

			msg = new byte[] { (byte)0x6b,(byte)0xc1,(byte)0xbe,(byte)0xe2,(byte)0x2e,(byte)0x40,(byte)0x9f,(byte)0x96,(byte)0xe9,(byte)0x3d,(byte)0x7e,(byte)0x11,(byte)0x73,(byte)0x93,(byte)0x17,(byte)0x2a,(byte)0xae,(byte)0x2d,(byte)0x8a,(byte)0x57,(byte)0x1e,(byte)0x03,(byte)0xac,(byte)0x9c,(byte)0x9e,(byte)0xb7,(byte)0x6f,(byte)0xac,(byte)0x45,(byte)0xaf,(byte)0x8e,(byte)0x51,(byte)0x30,(byte)0xc8,(byte)0x1c,(byte)0x46,(byte)0xa3,(byte)0x5c,(byte)0xe4,(byte)0x11 };
			ref = new byte[] { (byte)0xdf,(byte)0xa6,(byte)0x67,(byte)0x47,(byte)0xde,(byte)0x9a,(byte)0xe6,(byte)0x30,(byte)0x30,(byte)0xca,(byte)0x32,(byte)0x61,(byte)0x14,(byte)0x97,(byte)0xc8,(byte)0x27 };
			mac = m.doFinal(msg);
			if (!Arrays.equals(ref, mac)) {
				throw new RuntimeException("AES CMAC Selftest failed.");
			}

			msg = new byte[] { (byte)0x6b,(byte)0xc1,(byte)0xbe,(byte)0xe2,(byte)0x2e,(byte)0x40,(byte)0x9f,(byte)0x96,(byte)0xe9,(byte)0x3d,(byte)0x7e,(byte)0x11,(byte)0x73,(byte)0x93,(byte)0x17,(byte)0x2a,(byte)0xae,(byte)0x2d,(byte)0x8a,(byte)0x57,(byte)0x1e,(byte)0x03,(byte)0xac,(byte)0x9c,(byte)0x9e,(byte)0xb7,(byte)0x6f,(byte)0xac,(byte)0x45,(byte)0xaf,(byte)0x8e,(byte)0x51,(byte)0x30,(byte)0xc8,(byte)0x1c,(byte)0x46,(byte)0xa3,(byte)0x5c,(byte)0xe4,(byte)0x11,(byte)0xe5,(byte)0xfb,(byte)0xc1,(byte)0x19,(byte)0x1a,(byte)0x0a,(byte)0x52,(byte)0xef,(byte)0xf6,(byte)0x9f,(byte)0x24,(byte)0x45,(byte)0xdf,(byte)0x4f,(byte)0x9b,(byte)0x17,(byte)0xad,(byte)0x2b,(byte)0x41,(byte)0x7b,(byte)0xe6,(byte)0x6c,(byte)0x37,(byte)0x10 };
			ref = new byte[] { (byte)0x51,(byte)0xf0,(byte)0xbe,(byte)0xbf,(byte)0x7e,(byte)0x3b,(byte)0x9d,(byte)0x92,(byte)0xfc,(byte)0x49,(byte)0x74,(byte)0x17,(byte)0x79,(byte)0x36,(byte)0x3c,(byte)0xfe };
			mac = m.doFinal(msg);
			if (!Arrays.equals(ref, mac)) {
				throw new RuntimeException("AES CMAC Selftest failed.");
			}
		}
	}
}
