/*
 * Copyright (c) 2016 CardContact Systems GmbH, Minden, Germany.
 *
 * Redistribution and use in source (source code) and binary (object code)
 * forms, with or without modification, are permitted provided that the
 * following conditions are met:
 * 1. Redistributed source code must retain the above copyright notice, this
 * list of conditions and the disclaimer below.
 * 2. Redistributed object code must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation
 * and/or other materials provided with the distribution.
 * 3. The name of CardContact may not be used to endorse or promote products derived
 * from this software or in any other form without specific prior written
 * permission from CardContact.
 * 4. Redistribution of any modified code must be labeled "Code derived from
 * the original OpenCard Framework".
 *
 * THIS SOFTWARE IS PROVIDED BY CardContact "AS IS" FREE OF CHARGE. CardContact SHALL NOT BE
 * LIABLE FOR INFRINGEMENTS OF THIRD PARTIES RIGHTS BASED ON THIS SOFTWARE.  ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  CardContact DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THIS
 * SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR THAT THE OPERATION OF IT WILL
 * BE UNINTERRUPTED OR ERROR-FREE.  IN NO EVENT, UNLESS REQUIRED BY APPLICABLE
 * LAW, SHALL CardContact BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  ALSO, CardContact IS UNDER NO OBLIGATION
 * TO MAINTAIN, CORRECT, UPDATE, CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS
 * SOFTWARE.
 */

package de.cardcontact.opencard.terminal.remoteterminal;

import java.util.Properties;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.cardcontact.opencard.service.isocard.IsoConstants;
import de.cardcontact.opencard.service.remoteclient.RemoteCardSpec;
import de.cardcontact.opencard.service.remoteclient.RemoteProtocolScript;
import de.cardcontact.opencard.service.remoteclient.RemoteProtocolUnit;
import opencard.core.terminal.CardID;
import opencard.core.terminal.CardTerminal;
import opencard.core.terminal.CardTerminalException;
import opencard.core.terminal.CommandAPDU;
import opencard.core.terminal.CommunicationErrorException;
import opencard.core.terminal.ResponseAPDU;
import opencard.core.terminal.ScriptFailedException;
import opencard.core.terminal.SlotChannel;
import opencard.core.terminal.TerminalTimeoutException;

public class RemoteTerminal extends CardTerminal {

	private final static Logger logger = LoggerFactory.getLogger(RemoteTerminal.class);

	private static final int timeoutSingle = 2;		// Timeout for a single command in a sequence
	private static final int timeoutShort = 28;		// Timeout for a single command or last command in sequence
	private static final int timeoutLong = 300;		// Timeout for a long running command, like key generation
	private boolean freshConnect = true;
	private volatile boolean longPoll = false;
	private long lastPollingTimeout = -1;
	private int maxCAPDU = -1;
	private int maxRAPDU = -1;

	private CardID cardID = null;

	private LinkedBlockingQueue<RemoteProtocolScript> comQueue = new LinkedBlockingQueue<RemoteProtocolScript>(3);
	private LinkedBlockingQueue<RemoteProtocolScript> resQueue = new LinkedBlockingQueue<RemoteProtocolScript>(1);
	private RemoteProtocolScript current = null;



	public RemoteTerminal(String name) throws CardTerminalException {
		super(name, "RemoteTerminal", "");
		logger.debug("Created " + name);
		addSlots(1);
	}



	/**
	 * Check if the remote terminal has disappeared
	 *
	 * A connected terminal will poll for new C-APDUs using HTTP long polling with
	 * a period of 30 seconds. If the last polling sequence ended more that 5 second
	 * ago, then the reconnect in HTTP polling did not happen.
	 *
	 * @return true if long polling reconnect did not happen
	 */
	private boolean remoteDisappeared() {
		if (this.lastPollingTimeout != -1) {
			if (System.currentTimeMillis() - this.lastPollingTimeout > 5000) {
				logger.debug("Remote terminal disappeared at " + this.lastPollingTimeout);
				this.cardID = null;
				this.cardRemoved(0);
				return true;
			}
		}
		return false;
	}



	private void addRemoteProtocolUnit(RemoteProtocolUnit cmd) {
		if (this.current == null) {
			this.current = new RemoteProtocolScript();
		}
		this.current.add(cmd);
	}



	/**
	 * Transmit a command object and wait for a response object
	 *
	 * @param cmdObject the command object (i.e. CommandAPDU, Reset, Notify)
	 * @return the response object (i.e. ResponseAPDU, CardID, Notify)
	 * @throws CardTerminalException
	 */
	private RemoteProtocolScript transmit(int timeout) throws CardTerminalException {
		RemoteProtocolScript resObject;

		try {
			if (remoteDisappeared()) {
				throw new CardTerminalException("Remote terminal disappeared");
			}
			if (!comQueue.offer(this.current, 2, TimeUnit.SECONDS)) {
				throw new TerminalTimeoutException("Could not queue APDU to terminal. Remote card terminal probably disappeared", 1);
			}

			logger.debug("Waiting for R-APDU from remote terminal");

			resObject = resQueue.poll(timeout, TimeUnit.SECONDS);
			if (resObject == null) {
				throw new TerminalTimeoutException("The waiting time of " + timeout + " seconds for the response has expired.", timeout);
			}
		} catch (InterruptedException e) {
			throw new CardTerminalException(e.getMessage());
		} finally {
			this.current = null;
		}
		return resObject;
	}



	/**
	 * Poll for a command object. Used by remote connection.
	 * @return the command object (i.e. CommandAPDU, RemoteControl)
	 * @throws CardTerminalException
	 */
	public RemoteProtocolScript poll(int timeout) throws CardTerminalException {
		RemoteProtocolScript comObject;

		try {
			this.lastPollingTimeout = -1;
			this.longPoll = true;
			comObject = comQueue.poll(timeout, TimeUnit.SECONDS);
			if (comObject == null) {
				if (this.freshConnect) {
					throw new CommunicationErrorException("The waiting time of " + timeout + " seconds for the initial command apdu has expired.");
				}
				this.lastPollingTimeout = System.currentTimeMillis();
			}
			this.freshConnect = false;
		} catch (InterruptedException e) {
			throw new CardTerminalException(e.getMessage());
		} finally {
			this.longPoll = false;
		}

		return comObject;
	}



	/**
	 * Put response object into queue. Used by remote connection.
	 *
	 * @return the response object (i.e. ResponseAPDU, CardID, RemoteControl)
	 * @throws CardTerminalException
	 */
	public void put(RemoteProtocolScript resObject) {
		logger.debug("Placing R-APDU into queue");
		if (!resQueue.offer(resObject)) {
			logger.warn("Could not place message from remote into response queue. Queue stuck ?");
		}

		// Check if client closed the connection due
		for (RemoteProtocolUnit rpe : resObject.getRemoteProtocolUnits()) {
			if (rpe.isClosing()) {
				if (this.longPoll) {		// Add close for long polling request
					logger.debug("Cancel long polling");
					RemoteProtocolScript enc = new RemoteProtocolScript();
					enc.add(rpe);
					if (!comQueue.offer(enc)) {
						logger.warn("Could not place close message in command queue to end long polling request. Queue stuck ?");
					}
					cardRemoved(0);
				}
			}
		}
	}



	@Override
	public CardID getCardID(int slotID) throws CardTerminalException {
		return this.cardID;
	}



	/**
	 * Called when an initiation messages containing the card ATR is received.
	 *
	 * @param rcs
	 * @throws CardTerminalException
	 */
	public void initializeSession(RemoteCardSpec rcs) throws CardTerminalException {
		this.cardID = new CardID(this, 0, rcs.getCardID().getATR());
		this.maxCAPDU = rcs.getMaxCAPDU();
		this.maxRAPDU = rcs.getMaxRAPDU();
		this.freshConnect = true;
		this.comQueue.clear();
		this.resQueue.clear();
		this.cardInserted(0);
	}



	@Override
	protected Properties internalFeatures(Properties features) {
		if (this.maxCAPDU != -1) {
			features.put("maxCAPDUSize", String.valueOf(this.maxCAPDU));
		}
		if (this.maxRAPDU != -1) {
			features.put("maxRAPDUSize", String.valueOf(this.maxRAPDU));
		}
		return features;
	}



	@Override
	public boolean isCardPresent(int slotID) throws CardTerminalException {
		return cardID != null;
	}



	@Override
	public void open() throws CardTerminalException {
		logger.debug("[open] open");
	}



	@Override
	public void close() throws CardTerminalException {
		logger.debug("[close] close");

		if (this.cardID != null) {
			internalCloseSlotChannel(null);
		}
	}



	@Override
	protected CardID internalReset(int slot, int ms) throws CardTerminalException {

		addRemoteProtocolUnit(new RemoteProtocolUnit(RemoteProtocolUnit.Action.RESET));
		RemoteProtocolScript rpe = transmit(timeoutShort);

		for (RemoteProtocolUnit rpu: rpe.getRemoteProtocolUnits()) {
			switch(rpu.getAction()) {
			case RESET:
				this.cardID = new CardID(this, 0, ((RemoteCardSpec)rpu.getPayload()).getCardID().getATR());
				break;
			case CLOSE:
				throw new CommunicationErrorException(rpu.getMessage());
			default:
				break;
			}
		}

		return this.cardID;
	}



	/**
	 * Process a single Command APDU and return a Response APDU.
	 *
	 * This implementation support queueing of APDUs to combine subsequent
	 * APDUs into a script. If the queueable flag is set, then a dummy 9000
	 * is returned and the command is added to a work queue. If queueable is
	 * not set, then any previously queued commands and the current command are
	 * send to the client.
	 *
	 * The client can respond with either a list of response APDUs or just the last
	 * response APDU, in which case the number of executed commands must match the
	 * number command APDUs send. If a list of response APDUs is send, then the number
	 * of responses must match the number of commands and no intermediate status word
	 * may be other than 9000.
	 *
	 * @param slot the slot channel (deprecated)
	 * @param capdu the command APDU
	 * @param ms timeout (deprecated)
	 */
	@Override
	protected ResponseAPDU internalSendAPDU(int slot, CommandAPDU capdu, int ms)
			throws CardTerminalException {

		addRemoteProtocolUnit(new RemoteProtocolUnit(capdu));
		int cmds = this.current.getExecutedCommands() + 1;
		this.current.setExecutedCommands(cmds);

		ResponseAPDU rsp = null;

		if (capdu.isQueueable()) {
			rsp = new ResponseAPDU(new byte[] { (byte)0x90, 0x00 } );
			rsp.setQueued(true);
			return rsp;
		}

		int timeout = timeoutShort + cmds * timeoutSingle;
		if ((capdu.getByte(1) == IsoConstants.INS_GENERATE_KEYPAIR) || (capdu.getByte(1) == (IsoConstants.INS_INSTALL & 0xFF))) {
			timeout = timeoutLong;
		}

		RemoteProtocolScript rpe = transmit(timeout);

		int rsps = rpe.getExecutedCommands();
		int step = 0;
		int lastsw = IsoConstants.RC_OK;

		for (RemoteProtocolUnit rpu: rpe.getRemoteProtocolUnits()) {
			switch(rpu.getAction()) {
			case APDU:
				rsp = (ResponseAPDU)rpu.getPayload();
				// The client may returns a list of SW, one for each executed command
				// In that case only the last command may return anything other than 9000
				if (lastsw != IsoConstants.RC_OK) {
					logger.debug("Intermediate response APDU is not 9000");
					throw new ScriptFailedException(step, lastsw);
				}
				lastsw = rsp.sw();
				step++;
				break;
			case CLOSE:
				throw new CommunicationErrorException(rpu.getMessage());
			default:
				break;
			}
		}

		if (rsp == null) {
			throw new CardTerminalException("Response APDU missing");
		}

		if (step == 1) {	// Single response
			if ((cmds > 1) && (rsps < cmds)) {
				logger.debug("Did not execute all command APDUs (" + rsps + " of " + cmds + ")");
				throw new ScriptFailedException(step, lastsw);
			}
		} else {			// List of response APDUs returned
			if (step != cmds) {
				logger.debug("Did not execute all command APDUs (" + step + " of " + cmds + ")");
				throw new ScriptFailedException(step, lastsw);
			}
		}

		return rsp;
	}



	@Override
	protected void internalCloseSlotChannel(SlotChannel sc) {
		try	{
			if (this.current != null) {
				if (!comQueue.offer(this.current, 1, TimeUnit.SECONDS)) {
					logger.debug("Close message could not be added to queue");
				}
				this.current = null;
			}

			RemoteProtocolScript enc = new RemoteProtocolScript();
			enc.setClosing();
			if (!comQueue.offer(enc, 1, TimeUnit.SECONDS)) {
				logger.debug("Close message could not be added to queue");
			}
		} catch (InterruptedException e) {
			logger.debug("Adding close message interrupted");
		}
		cardID = null;
	}



	/**
	 * Send an asynchronous notification to the client
	 *
	 * @param id the integer status code, < 0 denotes an error, 0 denotes completion
	 * @param message the message to be send
	 * @param ttc the estimated time in seconds to complete this interaction
	 * @throws CardTerminalException
	 */
	public void sendNotification(int id, String message, int ttc) throws CardTerminalException {
		addRemoteProtocolUnit(new RemoteProtocolUnit(RemoteProtocolUnit.Action.NOTIFY, id, message, ttc));
	}



	/**
	 * Send an asynchronous notification to the client
	 *
	 * @param id the integer status code, < 0 denotes an error, 0 denotes completion
	 * @param message the message to be send
	 * @throws CardTerminalException
	 */
	public void sendNotification(int id, String message) throws CardTerminalException {
		sendNotification(id, message, 0);
	}
}
